#!/usr/bin/env python3
# nicked off hamster-service

import dbus
import dbus.service
import os.path
import subprocess

from dbus.mainloop.glib import DBusGMainLoop
from gi.repository import GLib as glib

import hamster


DBusGMainLoop(set_as_default=True)
loop = glib.MainLoop()

if "org.gnome.Hamster.WindowServer" in dbus.SessionBus().list_names():
    print("Found hamster-window-service already running, exiting")
    quit()


# Legacy server. Still used by the shell-extension.
# New code _could_ access the org.gnome.Hamster.GUI actions directly,
# although the exact action names/data are subject to change.
# http://lazka.github.io/pgi-docs/Gio-2.0/classes/Application.html#Gio.Application
# > The actions are also exported on the session bus,
#   and GIO provides the Gio.DBusActionGroup wrapper
#   to conveniently access them remotely.
class WindowServer(dbus.service.Object):
    __dbus_object_path__ = "/org/gnome/Hamster/WindowServer"

    def __init__(self, loop):
        self.app = True
        self.mainloop = loop
        self.bus = dbus.SessionBus()
        bus_name = dbus.service.BusName("org.gnome.Hamster.WindowServer", bus=self.bus)
        dbus.service.Object.__init__(self, bus_name, self.__dbus_object_path__)

    @dbus.service.method("org.gnome.Hamster")
    def Quit(self):
        """Shutdown the service"""
        self.mainloop.quit()

    def _open_window(self, name):
        if hamster.installed:
            base_cmd = "hamster"
        else:
            # both scripts are at the same place in the source tree
            cwd = os.path.dirname(__file__)
            base_cmd = "python3 {}/hamster-cli.py".format(cwd)
        cmd = "{} {} &".format(base_cmd, name)
        subprocess.run(cmd, shell=True)

    @dbus.service.method("org.gnome.Hamster.WindowServer")
    def edit(self, id: int = 0):
        """Edit fact, given its id (int) in the database.

        For backward compatibility, if id is 0, create a brand new fact.
        """
        if id:
            # {:d} restrict the string format. Too many safeguards cannot hurt.
            self._open_window("edit {:d}".format(id))
        else:
            self._open_window("add")

    @dbus.service.method("org.gnome.Hamster.WindowServer")
    def overview(self):
        self._open_window("overview")

    @dbus.service.method("org.gnome.Hamster.WindowServer")
    def about(self):
        self._open_window("about")

    @dbus.service.method("org.gnome.Hamster.WindowServer")
    def preferences(self):
        self._open_window("prefs")


if __name__ == '__main__':
    from hamster.lib import i18n
    i18n.setup_i18n()

    glib.set_prgname(str(_("hamster-windows-service")))

    window_server = WindowServer(loop)

    print("hamster-window-service up")

    loop.run()



#!/usr/bin/env python3
# nicked off gwibber

import dbus
import dbus.service

from gi.repository import GLib as glib
from gi.repository import Gio as gio

import hamster
from hamster import logger as hamster_logger
from hamster.lib import i18n
i18n.setup_i18n()  # noqa: E402

from hamster.storage import db
from hamster.lib import datetime as dt
from hamster.lib import default_logger
from hamster.lib.dbus import (
    DBusMainLoop,
    fact_signature,
    from_dbus_date,
    from_dbus_fact,
    from_dbus_fact_json,
    from_dbus_range,
    to_dbus_fact,
    to_dbus_fact_json
)
from hamster.lib.fact import Fact, FactError

logger = default_logger(__file__)


DBusMainLoop(set_as_default=True)
loop = glib.MainLoop()

if "org.gnome.Hamster" in dbus.SessionBus().list_names():
    print("Found hamster-service already running, exiting")
    quit()


class Storage(db.Storage, dbus.service.Object):
    __dbus_object_path__ = "/org/gnome/Hamster"

    def __init__(self, loop):
        self.bus = dbus.SessionBus()
        bus_name = dbus.service.BusName("org.gnome.Hamster", bus=self.bus)


        dbus.service.Object.__init__(self, bus_name, self.__dbus_object_path__)
        db.Storage.__init__(self, unsorted_localized="")

        self.mainloop = loop

        self.__file = gio.File.new_for_path(__file__)
        self.__monitor = self.__file.monitor_file(gio.FileMonitorFlags.WATCH_MOUNTS | \
                                                  gio.FileMonitorFlags.SEND_MOVED,
                                                  None)
        self.__monitor.connect("changed", self._on_us_change)

    def run_fixtures(self):
        """we start with an empty database and then populate with default
           values. This way defaults can be localized!"""
        super(Storage, self).run_fixtures()

        # defaults
        defaults = [
            (_("Work"), [_("Reading news"),
                         _("Checking stocks"),
                         _("Super secret project X"),
                         _("World domination")]),
            (_("Day-to-day"), [_("Lunch"),
                               _("Watering flowers"),
                               _("Doing handstands")])
        ]
        if not self.get_categories():
            for category, activities in defaults:
                cat_id = self.add_category(category)
                for activity in activities:
                    self.add_activity(activity, cat_id)


    # stop service when we have been updated (will be brought back in next call)
    # anyway. should make updating simpler
    def _on_us_change(self, monitor, gio_file, event_uri, event):
        if event == gio.FileMonitorEvent.CHANGES_DONE_HINT:
            print("`{}` has changed. Quitting!".format(__file__))
            self.Quit()

    @dbus.service.signal("org.gnome.Hamster")
    def TagsChanged(self): pass
    def tags_changed(self):
        self.TagsChanged()

    @dbus.service.signal("org.gnome.Hamster")
    def FactsChanged(self): pass
    def facts_changed(self):
        self.FactsChanged()

    @dbus.service.signal("org.gnome.Hamster")
    def ActivitiesChanged(self): pass
    def activities_changed(self):
        self.ActivitiesChanged()

    @dbus.service.signal("org.gnome.Hamster")
    def ToggleCalled(self): pass
    def toggle_called(self):
        self.toggle_called()

    def dispatch_overwrite(self):
        self.TagsChanged()
        self.FactsChanged()
        self.ActivitiesChanged()

    @dbus.service.method("org.gnome.Hamster")
    def Quit(self):
        """
        Shutdown the service
        example:
            import dbus
            obj = dbus.SessionBus().get_object("org.gnome.Hamster", "/org/gnome/Hamster")
            service = dbus.Interface(obj, "org.gnome.Hamster")
            service.Quit()
        """
        #log.logger.info("Hamster Service is being shutdown")
        self.mainloop.quit()


    @dbus.service.method("org.gnome.Hamster")
    def Toggle(self):
        """Toggle visibility of the main application window.
           If several instances are available, it will toggle them all.
        """
        #log.logger.info("Hamster Service is being shutdown")
        self.ToggleCalled()

    # facts
    @dbus.service.method("org.gnome.Hamster", in_signature='siib', out_signature='i')
    def AddFact(self, fact_str, start_time, end_time, temporary):
        """Add fact specified by a string.

        If the parsed fact has no start, then now is used.
        To fully use the hamster fact parser, as on the cmdline,
        just pass 0 for start_time and end_time.

        Args:
            fact_str (str): string to be parsed.
            start_time (int): Start datetime ovveride timestamp (ignored if 0).
                              -1 means None.
            end_time (int): datetime ovveride timestamp (ignored if 0).
                            -1 means None.
            #temporary (boolean): historical mystery, ignored, but needed to
                                 keep the method signature stable.
                                 Do not forget to pass something (e.g. False)!
        Returns:
            fact id (int), 0 means failure.

        Note: see datetime.utcfromtimestamp documentation
              for the precise meaning of timestamps.
        """
        fact = Fact.parse(fact_str)

        # default value if none found
        if not fact.start_time:
            fact.start_time = dt.datetime.now()

        if start_time == -1:
            fact.start_time = None
        elif start_time != 0:
            fact.start_time = dt.datetime.utcfromtimestamp(start_time)

        if end_time == -1:
            fact.end_time = None
        elif end_time != 0:
            fact.end_time = dt.datetime.utcfromtimestamp(end_time)

        return self.add_fact(fact)


    @dbus.service.method("org.gnome.Hamster", in_signature='s', out_signature='i')
    def AddFactJSON(self, dbus_fact):
        """Add fact given in JSON format.

        This is the preferred method if the fact fields are known separately,
        as activity, category, description and tags are passed "verbatim".
        Only datetimes are interpreted
        (2020-01-20: JSON does not know datetimes).

        Args:
            dbus_fact (str): fact in JSON format (cf. from_dbus_fact_json).

        Returns:
            fact id (int), 0 means failure.
        """
        fact = from_dbus_fact_json(dbus_fact)
        return self.add_fact(fact)


    @dbus.service.method("org.gnome.Hamster",
                         in_signature="si",
                         out_signature='bs')
    def CheckFact(self, dbus_fact, dbus_default_day):
        """Check fact validity.

        Useful to determine in advance whether the fact
        can be included in the database.

        Args:
            dbus_fact (str): fact in JSON format (cf. AddFactJSON)

        Returns:
            success (boolean): True upon success.
            message (str): what's wrong.
        """

        fact = from_dbus_fact_json(dbus_fact)
        dd = from_dbus_date(dbus_default_day)
        try:
            self.check_fact(fact, default_day=dd)
            success = True
            message = ""
        except FactError as error:
            success = False
            message = str(error)
        return success, message


    @dbus.service.method("org.gnome.Hamster",
                         in_signature='i',
                         out_signature=fact_signature)
    def GetFact(self, fact_id):
        """Get fact by id. For output format see GetFacts"""
        fact = self.get_fact(fact_id)
        return to_dbus_fact(fact)


    @dbus.service.method("org.gnome.Hamster",
                         in_signature='i',
                         out_signature="s")
    def GetFactJSON(self, fact_id):
        """Get fact by id.

        Return fact in JSON format (cf. to_dbus_fact_json)
        """
        fact = self.get_fact(fact_id)
        return to_dbus_fact_json(fact)


    @dbus.service.method("org.gnome.Hamster", in_signature='isiib', out_signature='i')
    def UpdateFact(self, fact_id, fact, start_time, end_time, temporary):
        start_time = start_time or None
        if start_time:
            start_time = dt.datetime.utcfromtimestamp(start_time)

        end_time = end_time or None
        if end_time:
            end_time = dt.datetime.utcfromtimestamp(end_time)
        return self.update_fact(fact_id, fact, start_time, end_time, temporary)


    @dbus.service.method("org.gnome.Hamster",
                         in_signature='is',
                         out_signature='i')
    def UpdateFactJSON(self, fact_id, dbus_fact):
        """Update fact.

        Args:
            fact_id (int): fact id in the database.
            dbus_fact (str): new fact content, in JSON format.
        Returns:
            int: new id (0 means failure)
        """
        fact = from_dbus_fact_json(dbus_fact)
        return self.update_fact(fact_id, fact)


    @dbus.service.method("org.gnome.Hamster", in_signature='i')
    def StopTracking(self, end_time):
        """Stops tracking the current activity"""
        end_time = end_time or None
        if end_time:
            end_time = dt.datetime.utcfromtimestamp(end_time)
        return self.stop_tracking(end_time)


    @dbus.service.method("org.gnome.Hamster", in_signature='i')
    def RemoveFact(self, fact_id):
        """Remove fact from storage by it's ID"""
        return self.remove_fact(fact_id)


    @dbus.service.method("org.gnome.Hamster",
                         in_signature='uus',
                         out_signature='a{}'.format(fact_signature))
    def GetFacts(self, start_date, end_date, search_terms):
        """Gets facts between the day of start_date and the day of end_date.
        Parameters:
        i start_date: Seconds since epoch (timestamp). Use 0 for today
        i end_date: Seconds since epoch (timestamp). Use 0 for today
        s search_terms: Bleh. If starts with "not ", the search terms will be reversed
        Returns an array of D-Bus fact structures.

        Legacy. To be superceded by GetFactsJSON at some point.
        """
        #TODO: Assert start > end ?
        start = dt.date.today()
        if start_date:
            start = dt.datetime.utcfromtimestamp(start_date).date()

        end = None
        if end_date:
            end = dt.datetime.utcfromtimestamp(end_date).date()

        return [to_dbus_fact(fact) for fact in self.get_facts(start, end, search_terms)]


    @dbus.service.method("org.gnome.Hamster",
                         in_signature='ss',
                         out_signature='as')
    def GetFactsJSON(self, dbus_range, search_terms):
        """Gets facts between the day of start and the day of end.

        Args:
            dbus_range (str): same format as on the command line.
                              (cf. dt.Range.parse)
            search_terms (str): If starts with "not ",
                                the search terms will be reversed
        Return:
            array of D-Bus facts in JSON format.
            (cf. to_dbus_fact_json)

        This will be the preferred way to get facts.
        """
        range = from_dbus_range(dbus_range)
        return [to_dbus_fact_json(fact)
                for fact in self.get_facts(range, search_terms=search_terms)]


    @dbus.service.method("org.gnome.Hamster", out_signature='a{}'.format(fact_signature))
    def GetTodaysFacts(self):
        """Gets facts of today,
           respecting hamster midnight. See GetFacts for return info.

           Legacy, to be superceded by GetTodaysFactsJSON at some point.
        """
        return [to_dbus_fact(fact) for fact in self.get_todays_facts()]


    @dbus.service.method("org.gnome.Hamster", out_signature='as')
    def GetTodaysFactsJSON(self):
        """Gets facts of the current hamster day.

        Return an array of facts in JSON format.
        """
        return [to_dbus_fact_json(fact) for fact in self.get_todays_facts()]


    # categories
    @dbus.service.method("org.gnome.Hamster", in_signature='s', out_signature = 'i')
    def AddCategory(self, name):
        return self.add_category(name)

    @dbus.service.method("org.gnome.Hamster", in_signature='s', out_signature='i')
    def GetCategoryId(self, category):
        return self.get_category_id(category)

    @dbus.service.method("org.gnome.Hamster", in_signature='is')
    def UpdateCategory(self, id, name):
        self.update_category(id, name)

    @dbus.service.method("org.gnome.Hamster", in_signature='i')
    def RemoveCategory(self, id):
        self.remove_category(id)

    @dbus.service.method("org.gnome.Hamster", out_signature='a(is)')
    def GetCategories(self):
        return [(category['id'], category['name']) for category in self.get_categories()]


    # activities
    @dbus.service.method("org.gnome.Hamster", in_signature='si', out_signature = 'i')
    def AddActivity(self, name, category_id):
        return self.add_activity(name, category_id)

    @dbus.service.method("org.gnome.Hamster", in_signature='isi')
    def UpdateActivity(self, id, name, category_id):
        self.update_activity(id, name, category_id)

    @dbus.service.method("org.gnome.Hamster", in_signature='i')
    def RemoveActivity(self, id):
        return self.remove_activity(id)

    @dbus.service.method("org.gnome.Hamster", in_signature='i', out_signature='a(isis)')
    def GetCategoryActivities(self, category_id):
        return [(row['id'],
                 row['name'],
                 row['category_id'],
                 row['category'] or '') for row in
                      self.get_category_activities(category_id = category_id)]


    @dbus.service.method("org.gnome.Hamster", in_signature='s', out_signature='a(ss)')
    def GetActivities(self, search = ""):
        return [(row['name'], row['category'] or '') for row in self.get_activities(search)]


    @dbus.service.method("org.gnome.Hamster", in_signature='ii', out_signature = 'b')
    def ChangeCategory(self, id, category_id):
        return self.change_category(id, category_id)


    @dbus.service.method("org.gnome.Hamster", in_signature='sib', out_signature='a{sv}')
    def GetActivityByName(self, activity, category_id, resurrect = True):
        category_id = category_id or None
        if activity:
            return dict(self.get_activity_by_name(activity, category_id, resurrect) or {})
        else:
            return {}

    # tags
    @dbus.service.method("org.gnome.Hamster", in_signature='b', out_signature='a(isb)')
    def GetTags(self, only_autocomplete):
        return [(tag['id'], tag['name'], tag['autocomplete']) for tag in self.get_tags(only_autocomplete)]


    @dbus.service.method("org.gnome.Hamster", in_signature='as', out_signature='a(isb)')
    def GetTagIds(self, tags):
        return [(tag['id'], tag['name'], tag['autocomplete']) for tag in self.get_tag_ids(tags)]


    @dbus.service.method("org.gnome.Hamster", in_signature='s')
    def SetTagsAutocomplete(self, tags):
        self.update_autocomplete_tags(tags)


    @dbus.service.method("org.gnome.Hamster", out_signature='s')
    def Version(self):
        return hamster.__version__


if __name__ == '__main__':
    import argparse
    parser = argparse.ArgumentParser(description="Hamster time tracker D-Bus service")

    # cf. https://stackoverflow.com/a/28611921/3565696
    parser.add_argument("--log", dest="log_level",
                        choices=('DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL'),
                        default='WARNING',
                        help="Set the logging level (default: %(default)s)")

    args = parser.parse_args()

    # logger for current script
    logger.setLevel(args.log_level)
    # hamster_logger for the rest
    hamster_logger.setLevel(args.log_level)

    print("hamster-service up")
    storage = Storage(loop)
    loop.run()



import gi
gi.require_version('Gtk', '3.0')  # noqa: E402
gi.require_version('PangoCairo', '1.0')  # noqa: E402
# for some reason performance is improved by importing Gtk early
from gi.repository import Gtk as gtk

from hamster.lib import default_logger


logger = default_logger(__name__)

try:
    # defs.py is created by waf from defs.py.in
    from hamster import defs
    __version__ = defs.VERSION
    installed = True
except ImportError:
    # if defs is not there, we are running from sources
    from subprocess import getstatusoutput
    rc, output = getstatusoutput("git describe --tags --always --dirty=+")
    __version__ = "3.0.2" if rc else "{} (uninstalled)".format(output)
    installed = False
    del getstatusoutput, rc, output

# cleanup namespace
del default_logger
del gtk  # performance is retained



import logging
logger = logging.getLogger(__name__)   # noqa: E402

import re

from hamster.lib import datetime as dt


# separator between times and activity
ACTIVITY_SEPARATOR = "\s+"


# match #tag followed by any space or # that will be ignored
# tag must not contain '#' or ','
tag_re = re.compile(r"""
    \#          # hash character
    (?P<tag>
        [^#,]+  # (anything but hash or comma)
    )
    \s*         # maybe spaces
                # forbid double comma (tag can not be before the tags barrier):
    ,?          # single comma (or none)
    \s*         # maybe space
    $           # end of text
""", flags=re.VERBOSE)

tags_separator = re.compile(r"""
    (,{0,2})    # 0, 1 or 2 commas
    \s*         # maybe spaces
    $           # end of text
""", flags=re.VERBOSE)


def parse_fact(text, range_pos="head", default_day=None, ref="now"):
    """Extract fact fields from the string.

    Returns found fields as a dict.

    Tentative syntax (not accurate):
    start [- end_time] activity[@category][,, description][,,]{ #tag}
    According to the legacy tests, # were allowed in the description
    """

    res = {}

    text = text.strip()
    if not text:
        return res

    # datetimes
    # force at least a space to avoid matching 10.00@cat
    (start, end), remaining_text = dt.Range.parse(text, position=range_pos,
                                                   separator=ACTIVITY_SEPARATOR,
                                                   default_day=default_day)
    res["start_time"] = start
    res["end_time"] = end

    # tags
    # Need to start from the end, because
    # the description can hold some '#' characters
    tags = []
    while True:
        # look for tags separators
        # especially the tags barrier
        m = re.search(tags_separator, remaining_text)
        remaining_text = remaining_text[:m.start()]
        if m.group(1) == ",,":
            # tags  barrier found
            break

        # look for tag
        m = re.search(tag_re, remaining_text)
        if m:
            tag = m.group('tag').strip()
            # strip the matched string (including #)
            remaining_text = remaining_text[:m.start()]
            tags.append(tag)
        else:
            # no tag
            break

    # put tags back in input order
    res["tags"] = list(reversed(tags))

    # description
    # first look for double comma (description hard left boundary)
    head, sep, description = remaining_text.partition(",,")
    res["description"] = description.strip()
    remaining_text = head.strip()

    # activity
    split = remaining_text.rsplit('@', maxsplit=1)
    activity = split[0]
    category = split[1] if len(split) > 1 else ""
    res["activity"] = activity
    res["category"] = category

    return res



#!/usr/bin/env python
# encoding: utf-8
# Thomas Nagy 2009-2018 (ita)

"""
Detects the Intel C++ compiler
"""

import sys
from waflib.Tools import ccroot, ar, gxx
from waflib.Configure import conf

@conf
def find_icpc(conf):
	"""
	Finds the program icpc, and execute it to ensure it really is icpc
	"""
	cxx = conf.find_program('icpc', var='CXX')
	conf.get_cc_version(cxx, icc=True)
	conf.env.CXX_NAME = 'icc'

def configure(conf):
	conf.find_icpc()
	conf.find_ar()
	conf.gxx_common_flags()
	conf.gxx_modifier_platform()
	conf.cxx_load_tools()
	conf.cxx_add_flags()
	conf.link_add_flags()




#!/usr/bin/env python3
# - coding: utf-8 -

# Copyright (C) 2010 Matías Ribecky <matias at mribecky.com.ar>
# Copyright (C) 2010-2012 Toms Bauģis <toms.baugis@gmail.com>
# Copyright (C) 2012 Ted Smith <tedks at cs.umd.edu>

# This file is part of Project Hamster.

# Project Hamster is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Project Hamster is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Project Hamster.  If not, see <http://www.gnu.org/licenses/>.

'''A script to control the applet from the command line.'''

import sys, os
import argparse
import re

import gi
gi.require_version('Gdk', '3.0')  # noqa: E402
gi.require_version('Gtk', '3.0')  # noqa: E402
from gi.repository import GLib as glib
from gi.repository import Gdk as gdk
from gi.repository import Gtk as gtk
from gi.repository import Gio as gio
from gi.repository import GLib as glib

import hamster

from hamster import client, reports
from hamster import logger as hamster_logger
from hamster.about import About
from hamster.edit_activity import CustomFactController
from hamster.overview import Overview
from hamster.preferences import PreferencesEditor
from hamster.lib import default_logger, stuff
from hamster.lib import datetime as dt
from hamster.lib.fact import Fact


logger = default_logger(__file__)


def word_wrap(line, max_len):
    """primitive word wrapper"""
    lines = []
    cur_line, cur_len = "", 0
    for word in line.split():
        if len("%s %s" % (cur_line, word)) < max_len:
            cur_line = ("%s %s" % (cur_line, word)).strip()
        else:
            if cur_line:
                lines.append(cur_line)
            cur_line = word
    if cur_line:
        lines.append(cur_line)
    return lines


def fact_dict(fact_data, with_date):
    fact = {}
    if with_date:
        fmt = '%Y-%m-%d %H:%M'
    else:
        fmt = '%H:%M'

    fact['start'] = fact_data.start_time.strftime(fmt)
    if fact_data.end_time:
        fact['end'] = fact_data.end_time.strftime(fmt)
    else:
        end_date = dt.datetime.now()
        fact['end'] = ''

    fact['duration'] = fact_data.delta.format()

    fact['activity'] = fact_data.activity
    fact['category'] = fact_data.category
    if fact_data.tags:
        fact['tags'] = ' '.join('#%s' % tag for tag in fact_data.tags)
    else:
        fact['tags'] = ''

    fact['description'] = fact_data.description

    return fact


class Hamster(gtk.Application):
    """Hamster gui.

    Actions should eventually be accessible via Gio.DBusActionGroup
    with the 'org.gnome.Hamster.GUI' id.
    but that is still experimental, the actions API is subject to change.
    Discussion with "external" developers welcome !
    The separate dbus org.gnome.Hamster.WindowServer
    is still the stable recommended way to show windows for now.
    """

    def __init__(self):
        # inactivity_timeout: How long (ms) the service should stay alive
        #                     after all windows have been closed.
        gtk.Application.__init__(self,
                                 application_id="org.gnome.Hamster.GUI",
                                 #inactivity_timeout=10000,
                                 register_session=True)

        self.about_controller = None  # 'about' window controller
        self.fact_controller = None  # fact window controller
        self.overview_controller = None  # overview window controller
        self.preferences_controller = None  # settings window controller

        self.connect("startup", self.on_startup)
        self.connect("activate", self.on_activate)

        # we need them before the startup phase
        # so register/activate_action work before the app is ran.
        # cf. https://gitlab.gnome.org/GNOME/glib/blob/master/gio/tests/gapplication-example-actions.c
        self.add_actions()

    def add_actions(self):
        # most actions have no parameters
        # for type "i", use Variant.new_int32() and .get_int32() to pack/unpack
        for name in ("about", "add", "clone", "edit", "overview", "preferences"):
            data_type = glib.VariantType("i") if name in ("edit", "clone") else None
            action = gio.SimpleAction.new(name, data_type)
            action.connect("activate", self.on_activate_window)
            self.add_action(action)

        action = gio.SimpleAction.new("quit", None)
        action.connect("activate", self.on_activate_quit)
        self.add_action(action)

    def on_activate(self, data=None):
        logger.debug("activate")
        if not self.get_windows():
            self.activate_action("overview")

    def on_activate_window(self, action=None, data=None):
        self._open_window(action.get_name(), data)

    def on_activate_quit(self, data=None):
        self.on_activate_quit()

    def on_startup(self, data=None):
        logger.debug("startup")
        # Must be the same as application_id. Won't be required with gtk4.
        glib.set_prgname(self.get_application_id())
        # localized name, but let's keep it simple.
        glib.set_application_name("Hamster")

    def _open_window(self, name, data=None):
        logger.debug("opening '{}'".format(name))

        if name == "about":
            if not self.about_controller:
                # silence warning "GtkDialog mapped without a transient parent"
                # https://stackoverflow.com/a/38408127/3565696
                _dummy = gtk.Window()
                self.about_controller = About(parent=_dummy)
                logger.debug("new About")
            controller = self.about_controller
        elif name in ("add", "clone", "edit"):
            if self.fact_controller:
                # Something is already going on, with other arguments, present it.
                # Or should we just discard the forgotten one ?
                logger.warning("Fact controller already active. Please close first.")
            else:
                fact_id = data.get_int32() if data else None
                self.fact_controller = CustomFactController(name, fact_id=fact_id)
                logger.debug("new CustomFactController")
            controller = self.fact_controller
        elif name == "overview":
            if not self.overview_controller:
                self.overview_controller = Overview()
                logger.debug("new Overview")
            controller = self.overview_controller
        elif name == "preferences":
            if not self.preferences_controller:
                self.preferences_controller = PreferencesEditor()
                logger.debug("new PreferencesEditor")
            controller = self.preferences_controller

        window = controller.window
        if window not in self.get_windows():
            self.add_window(window)
            logger.debug("window added")

        # Essential for positioning on wayland.
        # This should also select the correct window type if unset yet.
        # https://specifications.freedesktop.org/wm-spec/wm-spec-1.3.html
        if name != "overview" and self.overview_controller:
            window.set_transient_for(self.overview_controller.window)
            # so the dialog appears on top of the transient-for:
            window.set_type_hint(gdk.WindowTypeHint.DIALOG)
        else:
            # toplevel
            window.set_transient_for(None)

        controller.present()
        logger.debug("window presented")

    def present_fact_controller(self, action, fact_id=0):
        """Present the fact controller window to add, clone or edit a fact.

        Args:
            action (str): "add", "clone" or "edit"
        """
        assert action in ("add", "clone", "edit")
        if action in ("clone", "edit"):
            action_data = glib.Variant.new_int32(int(fact_id))
        else:
            action_data = None
        # always open dialogs through actions,
        # both for consistency, and to reduce the paths to test.
        app.activate_action(action, action_data)

class HamsterCli(object):
    """Command line interface."""

    def __init__(self):
        self.storage = client.Storage()


    def assist(self, *args):
        assist_command = args[0] if args else ""

        if assist_command == "start":
            hamster_client._activities(sys.argv[-1])
        elif assist_command == "export":
            formats = "html tsv xml ical".split()
            chosen = sys.argv[-1]
            formats = [f for f in formats if not chosen or f.startswith(chosen)]
            print("\n".join(formats))


    def toggle(self):
        self.storage.toggle()


    def start(self, *args):
        '''Start a new activity.'''
        if not args:
            print("Error: please specify activity")
            return 0

        fact = Fact.parse(" ".join(args), range_pos="tail")
        if fact.start_time is None:
            fact.start_time = dt.datetime.now()
        self.storage.check_fact(fact, default_day=dt.hday.today())
        id_ = self.storage.add_fact(fact)
        return id_


    def stop(self, *args):
        '''Stop tracking the current activity.'''
        self.storage.stop_tracking()


    def export(self, *args):
        args = args or []
        export_format, start_time, end_time = "html", None, None
        if args:
            export_format = args[0]
        (start_time, end_time), __ = dt.Range.parse(" ".join(args[1:]))

        start_time = start_time or dt.datetime.combine(dt.date.today(), dt.time())
        end_time = end_time or start_time.replace(hour=23, minute=59, second=59)
        facts = self.storage.get_facts(start_time, end_time)

        writer = reports.simple(facts, start_time.date(), end_time.date(), export_format)


    def _activities(self, search=""):
        '''Print the names of all the activities.'''
        if "@" in search:
            activity, category = search.split("@")
            for cat in self.storage.get_categories():
                if not category or cat['name'].lower().startswith(category.lower()):
                    print("{}@{}".format(activity, cat['name']))
        else:
            for activity in self.storage.get_activities(search):
                print(activity['name'])
                if activity['category']:
                    print("{}@{}".format(activity['name'], activity['category']))


    def activities(self, *args):
        '''Print the names of all the activities.'''
        search = args[0] if args else ""
        for activity in self.storage.get_activities(search):
            print("{}@{}".format(activity['name'], activity['category']))

    def categories(self, *args):
        '''Print the names of all the categories.'''
        for category in self.storage.get_categories():
            print(category['name'])


    def list(self, *times):
        """list facts within a date range"""
        (start_time, end_time), __ = dt.Range.parse(" ".join(times or []))

        start_time = start_time or dt.datetime.combine(dt.date.today(), dt.time())
        end_time = end_time or start_time.replace(hour=23, minute=59, second=59)
        self._list(start_time, end_time)


    def current(self, *args):
        """prints current activity. kinda minimal right now"""
        facts = self.storage.get_todays_facts()
        if facts and not facts[-1].end_time:
            print("{} {}".format(str(facts[-1]).strip(),
                                 facts[-1].delta.format(fmt="HH:MM")))
        else:
            print((_("No activity")))


    def search(self, *args):
        """search for activities by name and optionally within a date range"""
        args = args or []
        search = ""
        if args:
            search = args[0]

        (start_time, end_time), __ = dt.Range.parse(" ".join(args[1:]))

        start_time = start_time or dt.datetime.combine(dt.date.today(), dt.time())
        end_time = end_time or start_time.replace(hour=23, minute=59, second=59)
        self._list(start_time, end_time, search)


    def _list(self, start_time, end_time, search=""):
        """Print a listing of activities"""
        facts = self.storage.get_facts(start_time, end_time, search)


        headers = {'activity': _("Activity"),
                   'category': _("Category"),
                   'tags': _("Tags"),
                   'description': _("Description"),
                   'start': _("Start"),
                   'end': _("End"),
                   'duration': _("Duration")}


        # print date if it is not the same day
        print_with_date = start_time.date() != end_time.date()

        cols = 'start', 'end', 'duration', 'activity', 'category'


        widths = dict([(col, len(headers[col])) for col in cols])
        for fact in facts:
            fact = fact_dict(fact, print_with_date)
            for col in cols:
                widths[col] = max(widths[col], len(fact[col]))

        cols = ["{{{col}: <{len}}}".format(col=col, len=widths[col]) for col in cols]
        fact_line = " | ".join(cols)

        row_width = sum(val + 3 for val in list(widths.values()))

        print()
        print(fact_line.format(**headers))
        print("-" * min(row_width, 80))

        by_cat = {}
        for fact in facts:
            cat = fact.category or _("Unsorted")
            by_cat.setdefault(cat, dt.timedelta(0))
            by_cat[cat] += fact.delta

            pretty_fact = fact_dict(fact, print_with_date)
            print(fact_line.format(**pretty_fact))

            if pretty_fact['description']:
                for line in word_wrap(pretty_fact['description'], 76):
                    print("    {}".format(line))

            if pretty_fact['tags']:
                for line in word_wrap(pretty_fact['tags'], 76):
                    print("    {}".format(line))

        print("-" * min(row_width, 80))

        cats = []
        total_duration = dt.timedelta()
        for cat, duration in sorted(by_cat.items(), key=lambda x: x[1], reverse=True):
            cats.append("{}: {}".format(cat, duration.format()))
            total_duration += duration

        for line in word_wrap(", ".join(cats), 80):
            print(line)
        print("Total: ", total_duration.format())

        print()


    def version(self):
        print(hamster.__version__)


if __name__ == '__main__':
    from hamster.lib import i18n
    i18n.setup_i18n()

    usage = _(
"""
Actions:
    * add [activity [start-time [end-time]]]: Add an activity
    * stop: Stop tracking current activity.
    * list [start-date [end-date]]: List activities
    * search [terms] [start-date [end-date]]: List activities matching a search
      term
    * export [html|tsv|ical|xml] [start-date [end-date]]: Export activities with
      the specified format
    * current: Print current activity
    * activities: List all the activities names, one per line.
    * categories: List all the categories names, one per line.

    * overview / preferences / add / about: launch specific window

    * version: Show the Hamster version

Time formats:
    * 'YYYY-MM-DD hh:mm': If start-date is missing, it will default to today.
      If end-date is missing, it will default to start-date.
    * '-minutes': Relative time in minutes from the current date and time.
Note:
    * For list/search/export a "hamster day" starts at the time set in the
      preferences (default 05:00) and ends one minute earlier the next day.
      Activities are reported for each "hamster day" in the interval.

Example usage:
    hamster start bananas -20
        start activity 'bananas' with start time 20 minutes ago

    hamster search pancakes 2012-08-01 2012-08-30
        look for an activity matching terms 'pancakes` between 1st and 30st
        August 2012. Will check against activity, category, description and tags
""")

    hamster_client = HamsterCli()
    app = Hamster()
    logger.debug("app instanciated")

    import signal
    signal.signal(signal.SIGINT, signal.SIG_DFL) # gtk3 screws up ctrl+c

    parser = argparse.ArgumentParser(
        description="Time tracking utility",
        epilog=usage,
        formatter_class=argparse.RawDescriptionHelpFormatter)

    # cf. https://stackoverflow.com/a/28611921/3565696
    parser.add_argument("--log", dest="log_level",
                        choices=('DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL'),
                        default='WARNING',
                        help="Set the logging level (default: %(default)s)")
    parser.add_argument("action", nargs="?", default="overview")
    parser.add_argument('action_args', nargs=argparse.REMAINDER, default=[])

    args, unknown_args = parser.parse_known_args()

    # logger for current script
    logger.setLevel(args.log_level)
    # hamster_logger for the rest
    hamster_logger.setLevel(args.log_level)

    if not hamster.installed:
        logger.info("Running in devel mode")

    if args.action in ("start", "track"):
        action = "add"  # alias
    elif args.action == "prefs":
        # for backward compatibility
        action = "preferences"
    else:
        action = args.action

    if action in ("about", "add", "edit", "overview", "preferences"):
        if action == "add" and args.action_args:
            assert not unknown_args, "unknown options: {}".format(unknown_args)
            # directly add fact from arguments
            id_ = hamster_client.start(*args.action_args)
            assert id_ > 0, "failed to add fact"
            sys.exit(0)
        else:
            app.register()
            if action == "edit":
                assert len(args.action_args) == 1, (
                       "edit requires exactly one argument, got {}"
                       .format(args.action_args))
                id_ = int(args.action_args[0])
                assert id_ > 0, "received non-positive id : {}".format(id_)
                action_data = glib.Variant.new_int32(id_)
            else:
                action_data = None
            app.activate_action(action, action_data)
            run_args = [sys.argv[0]] + unknown_args
            logger.debug("run {}".format(run_args))
            status = app.run(run_args)
            logger.debug("app exited")
            sys.exit(status)
    elif hasattr(hamster_client, action):
        getattr(hamster_client, action)(*args.action_args)
    else:
        sys.exit(usage % {'prog': sys.argv[0]})



#!/usr/bin/env python2
import distutils
import os
from distutils.core import setup
from distutils import sysconfig

data_dir = os.path.join(sysconfig.get_python_lib(), "hamster", "data")

setup(name='hamster-sqlite',
      version='0.3',
      description='Minimal dependency nicely abstracted sqlite backend of hamster time tracker - lets you connect to your hamster db and do stuff in python',
      author='Toms Baugis',
      author_email='toms.baugis@gmail.com',
      url='https://github.com/projecthamster/hamster',
      package_dir = {'': 'src'},
      py_modules = ['hamster.storage', 'hamster.db', 'hamster.lib.__init__'],
      data_files=[(data_dir, ['data/hamster.db'])],
     )



import sys, os.path
# a convoluted line to add hamster module to absolute path
sys.path.insert(0, os.path.realpath(os.path.join(os.path.dirname(__file__), "../src")))

import datetime as pdt
import unittest
import re
from hamster.lib import datetime as dt
from hamster.lib.dbus import (
    to_dbus_fact,
    to_dbus_fact_json,
    to_dbus_range,
    from_dbus_fact,
    from_dbus_fact_json,
    from_dbus_range,
    )
from hamster.lib.fact import Fact


class TestFact(unittest.TestCase):

    def test_range(self):
        t1 = dt.datetime(2020, 1, 15, 13, 30)
        t2 = dt.datetime(2020, 1, 15, 15, 30)
        range = dt.Range(t1, t2)
        fact = Fact(range=range)
        self.assertEqual(fact.range.start, t1)
        self.assertEqual(fact.range.end, t2)
        fact = Fact(start=t1, end=t2)
        self.assertEqual(fact.range.start, t1)
        self.assertEqual(fact.range.end, t2)
        # backward compatibility (before v3.0)
        fact = Fact(start_time=t1, end_time=t2)
        self.assertEqual(fact.range.start, t1)
        self.assertEqual(fact.range.end, t2)


class TestFactParsing(unittest.TestCase):

    def test_plain_name(self):
        # plain activity name
        activity = Fact.parse("just a simple case with ütf-8")
        self.assertEqual(activity.activity, "just a simple case with ütf-8")
        assert activity.start_time is None
        assert activity.end_time is None
        assert not activity.category
        assert not activity.description

    def test_only_range(self):
        fact = Fact.parse("-20")
        assert not fact.activity
        fact = Fact.parse("-20 -10")
        assert not fact.activity

    def test_with_start_time(self):
        # with time
        activity = Fact.parse("12:35 with start time")
        self.assertEqual(activity.activity, "with start time")
        self.assertEqual(activity.start_time.strftime("%H:%M"), "12:35")

        #rest must be empty
        assert not activity.category
        assert activity.end_time is None
        assert not activity.description

    def test_with_start_and_end_time(self):
        # with time
        activity = Fact.parse("12:35-14:25 with start-end time")
        self.assertEqual(activity.activity, "with start-end time")
        self.assertEqual(activity.start_time.strftime("%H:%M"), "12:35")
        self.assertEqual(activity.end_time.strftime("%H:%M"), "14:25")

        #rest must be empty
        assert not activity.category
        assert not activity.description

    def test_category(self):
        # plain activity name
        activity = Fact.parse("just a simple case@hämster")
        self.assertEqual(activity.activity, "just a simple case")
        self.assertEqual(activity.category, "hämster")
        assert activity.start_time is None
        assert activity.end_time is None
        assert not activity.description

    def test_description(self):
        # plain activity name
        activity = Fact.parse("case,, with added descriptiön")
        self.assertEqual(activity.activity, "case")
        self.assertEqual(activity.description, "with added descriptiön")
        assert not activity.category
        assert activity.start_time is None
        assert activity.end_time is None
        assert not activity.category

    def test_tags(self):
        # plain activity name
        activity = Fact.parse("#case,, description with #hash,, #and, #some #tägs")
        self.assertEqual(activity.activity, "#case")
        self.assertEqual(activity.description, "description with #hash")
        self.assertEqual(set(activity.tags), set(["and", "some", "tägs"]))
        assert not activity.category
        assert activity.start_time is None
        assert activity.end_time is None

    def test_full(self):
        # plain activity name
        activity = Fact.parse("1225-1325 case@cat,, description #ta non-tag,, #tag #bäg")
        self.assertEqual(activity.start_time.strftime("%H:%M"), "12:25")
        self.assertEqual(activity.end_time.strftime("%H:%M"), "13:25")
        self.assertEqual(activity.activity, "case")
        self.assertEqual(activity.category, "cat")
        self.assertEqual(activity.description, "description #ta non-tag")
        self.assertEqual(set(activity.tags), set(["bäg", "tag"]))

    def test_copy(self):
        fact1 = Fact.parse("12:25-13:25 case@cat,, description #tag #bäg")
        fact2 = fact1.copy()
        self.assertEqual(fact1.start_time, fact2.start_time)
        self.assertEqual(fact1.end_time, fact2.end_time)
        self.assertEqual(fact1.activity, fact2.activity)
        self.assertEqual(fact1.category, fact2.category)
        self.assertEqual(fact1.description, fact2.description)
        self.assertEqual(fact1.tags, fact2.tags)
        fact3 = fact1.copy(activity="changed")
        self.assertEqual(fact3.activity, "changed")
        fact3 = fact1.copy(category="changed")
        self.assertEqual(fact3.category, "changed")
        fact3 = fact1.copy(description="changed")
        self.assertEqual(fact3.description, "changed")
        fact3 = fact1.copy(tags=["changed"])
        self.assertEqual(fact3.tags, ["changed"])

    def test_comparison(self):
        fact1 = Fact.parse("12:25-13:25 case@cat,, description #tag #bäg")
        fact2 = fact1.copy()
        self.assertEqual(fact1, fact2)
        fact2 = fact1.copy()
        fact2.activity = "abcd"
        self.assertNotEqual(fact1, fact2)
        fact2 = fact1.copy()
        fact2.category = "abcd"
        self.assertNotEqual(fact1, fact2)
        fact2 = fact1.copy()
        fact2.description = "abcd"
        self.assertNotEqual(fact1, fact2)
        fact2 = fact1.copy()
        fact2.range.start = fact1.range.start + dt.timedelta(minutes=1)
        self.assertNotEqual(fact1, fact2)
        fact2 = fact1.copy()
        fact2.range.end = fact1.range.end + dt.timedelta(minutes=1)
        self.assertNotEqual(fact1, fact2)
        # wrong order
        fact2 = fact1.copy()
        fact2.tags = ["bäg", "tag"]
        self.assertNotEqual(fact1, fact2)
        # correct order
        fact2 = fact1.copy()
        fact2.tags = ["tag", "bäg"]
        self.assertEqual(fact1, fact2)

    def test_decimal_in_activity(self):
        # cf. issue #270
        fact = Fact.parse("12:25-13:25 10.0@ABC,, Two Words #tag #bäg")
        self.assertEqual(fact.activity, "10.0")
        self.assertEqual(fact.category, "ABC")
        self.assertEqual(fact.description, "Two Words")
        # should not pick up a time here
        fact = Fact.parse("10.00@ABC,, Two Words #tag #bäg")
        self.assertEqual(fact.activity, "10.00")
        self.assertEqual(fact.category, "ABC")
        self.assertEqual(fact.description, "Two Words")

    def test_spaces(self):
        # cf. issue #114
        fact = Fact.parse("11:00 12:00 BPC-261 - Task title@Project#code")
        self.assertEqual(fact.activity, "BPC-261 - Task title")
        self.assertEqual(fact.category, "Project")
        self.assertEqual(fact.description, "")
        self.assertEqual(fact.tags, ["code"])
        # space between category and tag
        fact2 = Fact.parse("11:00 12:00 BPC-261 - Task title@Project #code")
        self.assertEqual(fact.serialized(), fact2.serialized())
        # empty fact
        fact3 = Fact()
        self.assertEqual(fact3.serialized(), "")

    def test_commas(self):
        fact = Fact.parse("11:00 12:00 activity, with comma@category,, description, with comma")
        self.assertEqual(fact.activity, "activity, with comma")
        self.assertEqual(fact.category, "category")
        self.assertEqual(fact.description, "description, with comma")
        self.assertEqual(fact.tags, [])
        fact = Fact.parse("11:00 12:00 activity, with comma@category,, description, with comma, #tag1, #tag2")
        self.assertEqual(fact.activity, "activity, with comma")
        self.assertEqual(fact.category, "category")
        self.assertEqual(fact.description, "description, with comma")
        self.assertEqual(fact.tags, ["tag1", "tag2"])
        fact = Fact.parse("11:00 12:00 activity, with comma@category,, description, with comma and #hash,, #tag1, #tag2")
        self.assertEqual(fact.activity, "activity, with comma")
        self.assertEqual(fact.category, "category")
        self.assertEqual(fact.description, "description, with comma and #hash")
        self.assertEqual(fact.tags, ["tag1", "tag2"])

    # ugly. Really need pytest
    def test_roundtrips(self):
        for start_time in (
            None,
            dt.time(12, 33),
            ):
            for end_time in (
                None,
                dt.time(13, 34),
                ):
                for activity in (
                    "activity",
                    "#123 with two #hash",
                    "activity, with comma",
                    "17.00 tea",
                    ):
                    for category in (
                        "",
                        "category",
                        ):
                        for description in (
                            "",
                            "description",
                            "with #hash",
                            "with, comma",
                            "with @at",
                            "multiline\ndescription",
                            ):
                            for tags in (
                                [],
                                ["single"],
                                ["with space"],
                                ["two", "tags"],
                                ["with @at"],
                                ):
                                start = dt.datetime.from_day_time(dt.hday.today(),
                                                                  start_time
                                                                  ) if start_time else None
                                end = dt.datetime.from_day_time(dt.hday.today(),
                                                                end_time
                                                                ) if end_time else None
                                if end and not start:
                                    # end without start is not parseable
                                    continue
                                fact = Fact(start_time=start,
                                            end_time=end,
                                            activity=activity,
                                            category=category,
                                            description=description,
                                            tags=tags)
                                for range_pos in ("head", "tail"):
                                    fact_str = fact.serialized(range_pos=range_pos)
                                    parsed = Fact.parse(fact_str, range_pos=range_pos)
                                    self.assertEqual(fact, parsed)
                                    self.assertEqual(parsed.range.start, fact.range.start)
                                    self.assertEqual(parsed.range.end, fact.range.end)
                                    self.assertEqual(parsed.activity, fact.activity)
                                    self.assertEqual(parsed.category, fact.category)
                                    self.assertEqual(parsed.description, fact.description)
                                    self.assertEqual(parsed.tags, fact.tags)


class TestDatetime(unittest.TestCase):
    def test_datetime_from_day_time(self):
        day = dt.date(2018, 8, 13)
        time = dt.time(23, 10)
        expected = dt.datetime(2018, 8, 13, 23, 10)  # 2018-08-13 23:10
        self.assertEqual(dt.datetime.from_day_time(day, time), expected)
        day = dt.date(2018, 8, 13)
        time = dt.time(0, 10)
        expected = dt.datetime(2018, 8, 14, 0, 10)  # 2018-08-14 00:10
        self.assertEqual(dt.datetime.from_day_time(day, time), expected)

    def test_format_timedelta(self):
        delta = dt.timedelta(minutes=10)
        self.assertEqual(delta.format("human"), "10min")
        delta = dt.timedelta(hours=5, minutes=0)
        self.assertEqual(delta.format("human"), "5h")
        delta = dt.timedelta(hours=5, minutes=10)
        self.assertEqual(delta.format("human"), "5h 10min")
        delta = dt.timedelta(hours=5, minutes=10)
        self.assertEqual(delta.format("HH:MM"), "05:10")

    def test_datetime_hday(self):
        date_time = dt.datetime(2018, 8, 13, 23, 10)  # 2018-08-13 23:10
        expected = dt.date(2018, 8, 13)
        self.assertEqual(date_time.hday(), expected)
        date_time = dt.datetime(2018, 8, 14, 0, 10)  # 2018-08-14 0:10
        expected = dt.date(2018, 8, 13)
        self.assertEqual(date_time.hday(), expected)
        today = dt.hday.today()
        self.assertEqual(type(today), dt.hday)

    def test_parse_date(self):
        date = dt.date.parse("2020-01-05")
        self.assertEqual(date, pdt.date(2020, 1, 5))

    def test_parse_time(self):
        self.assertEqual(dt.time.parse("9:01"), pdt.time(9, 1))
        self.assertEqual(dt.time.parse("9.01"), pdt.time(9, 1))
        self.assertEqual(dt.time.parse("12:01"), pdt.time(12, 1))
        self.assertEqual(dt.time.parse("12.01"), pdt.time(12, 1))
        self.assertEqual(dt.time.parse("1201"), pdt.time(12, 1))

    def test_parse_datetime(self):
        self.assertEqual(dt.datetime.parse("2020-01-05 9:01"), pdt.datetime(2020, 1, 5, 9, 1))

    def test_datetime_patterns(self):
        p = dt.datetime.pattern(1)
        s = "12:03"
        m = re.fullmatch(p, s, re.VERBOSE)
        time = dt.datetime._extract_datetime(m, d="date1", h="hour1", m="minute1", r="relative1",
                                              default_day=dt.hday.today())
        self.assertEqual(time.strftime("%H:%M"), "12:03")
        s = "2019-12-01 12:36"
        m = re.fullmatch(p, s, re.VERBOSE)
        time = dt.datetime._extract_datetime(m, d="date1", h="hour1", m="minute1", r="relative1")
        self.assertEqual(time.strftime("%Y-%m-%d %H:%M"), "2019-12-01 12:36")
        s = "-25"
        m = re.fullmatch(p, s, re.VERBOSE)
        relative = dt.datetime._extract_datetime(m, d="date1", h="hour1", m="minute1", r="relative1",
                                                  default_day=dt.hday.today())
        self.assertEqual(relative, dt.timedelta(minutes=-25))
        s = "2019-12-05"
        m = re.search(p, s, re.VERBOSE)
        self.assertEqual(m, None)

    def test_parse_datetime_range(self):
        # only match clean
        s = "10.00@cat"
        (start, end), rest = dt.Range.parse(s, position="head")
        self.assertEqual(start, None)
        self.assertEqual(end, None)
        s = "12:02"
        (start, end), rest = dt.Range.parse(s)
        self.assertEqual(start.strftime("%H:%M"), "12:02")
        self.assertEqual(end, None)
        s = "12:03 13:04"
        (start, end), rest = dt.Range.parse(s)
        self.assertEqual(start.strftime("%H:%M"), "12:03")
        self.assertEqual(end.strftime("%H:%M"), "13:04")
        s = "12:35 activity"
        (start, end), rest = dt.Range.parse(s, position="head")
        self.assertEqual(start.strftime("%H:%M"), "12:35")
        self.assertEqual(end, None)
        s = "2019-12-01 12:33 activity"
        (start, end), rest = dt.Range.parse(s, position="head")
        self.assertEqual(start.strftime("%Y-%m-%d %H:%M"), "2019-12-01 12:33")
        self.assertEqual(end, None)

        ref = dt.datetime(2019, 11, 29, 13, 55)  # 2019-11-29 13:55

        s = "-25 activity"
        (start, end), rest = dt.Range.parse(s, position="head", ref=ref)
        self.assertEqual(start.strftime("%Y-%m-%d %H:%M"), "2019-11-29 13:30")
        self.assertEqual(end, None)
        s = "+25 activity"
        (start, end), rest = dt.Range.parse(s, position="head", ref=ref)
        self.assertEqual(start.strftime("%Y-%m-%d %H:%M"), "2019-11-29 14:20")
        self.assertEqual(end, None)
        s = "-55 -25 activity"
        (start, end), rest = dt.Range.parse(s, position="head", ref=ref)
        self.assertEqual(start.strftime("%Y-%m-%d %H:%M"), "2019-11-29 13:00")
        self.assertEqual(end.strftime("%Y-%m-%d %H:%M"), "2019-11-29 13:30")
        s = "+25 +55 activity"
        (start, end), rest = dt.Range.parse(s, position="head", ref=ref)
        self.assertEqual(start.strftime("%Y-%m-%d %H:%M"), "2019-11-29 14:20")
        self.assertEqual(end.strftime("%Y-%m-%d %H:%M"), "2019-11-29 14:50")
        s = "-55 -120 activity"
        (start, end), rest = dt.Range.parse(s, position="head", ref=ref)
        self.assertEqual(start.strftime("%Y-%m-%d %H:%M"), "2019-11-29 13:00")
        self.assertEqual(end.strftime("%Y-%m-%d %H:%M"), "2019-11-29 11:55")
        s = "-50 20 activity"
        (start, end), rest = dt.Range.parse(s, position="head", ref=ref)
        self.assertEqual(start.strftime("%Y-%m-%d %H:%M"), "2019-11-29 13:05")
        self.assertEqual(end.strftime("%Y-%m-%d %H:%M"), "2019-11-29 13:25")

        s = "2019-12-05"  # single hamster day
        (start, end), rest = dt.Range.parse(s, ref=ref)
        just_before = start - dt.timedelta(seconds=1)
        just_after = end + dt.timedelta(seconds=1)
        self.assertEqual(just_before.hday(), pdt.date(2019, 12, 4))
        self.assertEqual(just_after.hday(), pdt.date(2019, 12, 6))
        s = "2019-12-05 2019-12-07"  # hamster days range
        (start, end), rest = dt.Range.parse(s, ref=ref)
        just_before = start - dt.timedelta(seconds=1)
        just_after = end + dt.timedelta(seconds=1)
        self.assertEqual(just_before.hday(), dt.date(2019, 12, 4))
        self.assertEqual(just_after.hday(), dt.date(2019, 12, 8))

        s = "14:30 - --"
        (start, end), rest = dt.Range.parse(s, ref=ref)
        self.assertEqual(start.strftime("%H:%M"), "14:30")
        self.assertEqual(end, None)

    def test_range(self):
        day = dt.hday(2020, 2, 2)
        time = dt.time(21, 20)
        base = dt.datetime.from_day_time(day, time)
        range = dt.Range(base, base + dt.timedelta(minutes=30))
        range_str = range.format(default_day=day)
        self.assertEqual(range_str, "21:20 - 21:50")
        range = dt.Range(None, base)
        range_str = range.format(default_day=day)
        self.assertEqual(range_str, "-- - 21:20")
        # issue #576
        start = dt.datetime(2020, 3, 8, 17, 7)
        end = dt.datetime(2020, 3, 8, 18, 6)
        range = dt.Range.from_start_end(start, end)
        self.assertEqual(range.start, start)
        self.assertEqual(range.end, end)
        # check passthrough
        range2 = dt.Range.from_start_end(range)
        self.assertEqual(range2.start, range.start)
        self.assertEqual(range2.end, range.end)

    def test_rounding(self):
        dt1 = dt.datetime(2019, 12, 31, hour=13, minute=14, second=10, microsecond=11)
        self.assertEqual(dt1.second, 0)
        self.assertEqual(dt1.microsecond, 0)
        self.assertEqual(str(dt1), "2019-12-31 13:14")

    def test_type_stability(self):
        dt1 = dt.datetime(2020, 1, 10, hour=13, minute=30)
        dt2 = dt.datetime(2020, 1, 10, hour=13, minute=40)
        delta = dt2 - dt1
        self.assertEqual(type(delta), dt.timedelta)
        _sum = dt1 + delta
        self.assertEqual(_sum, dt.datetime(2020, 1, 10, hour=13, minute=40))
        self.assertEqual(type(_sum), dt.datetime)
        _sub = dt1 - delta
        self.assertEqual(_sub, dt.datetime(2020, 1, 10, hour=13, minute=20))
        self.assertEqual(type(_sub), dt.datetime)

        opposite = - delta
        self.assertEqual(opposite, dt.timedelta(minutes=-10))
        self.assertEqual(type(opposite), dt.timedelta)
        _sum = delta + delta
        self.assertEqual(_sum, dt.timedelta(minutes=20))
        self.assertEqual(type(_sum), dt.timedelta)
        _sub = delta - delta
        self.assertEqual(_sub, dt.timedelta())
        self.assertEqual(type(_sub), dt.timedelta)

    def test_timedelta(self):
        delta = dt.timedelta(seconds=90)
        self.assertEqual(delta.total_minutes(), 1.5)


class TestDBus(unittest.TestCase):
    def test_round_trip(self):
        fact = Fact.parse("11:00 12:00 activity, with comma@category,, description, with comma #and #tags")
        dbus_fact = to_dbus_fact_json(fact)
        return_fact = from_dbus_fact_json(dbus_fact)
        self.assertEqual(return_fact, fact)

        dbus_fact = to_dbus_fact(fact)
        return_fact = from_dbus_fact(dbus_fact)
        self.assertEqual(return_fact, fact)

        fact = Fact.parse("11:00 activity")
        dbus_fact = to_dbus_fact_json(fact)
        return_fact = from_dbus_fact_json(dbus_fact)
        self.assertEqual(return_fact, fact)

        dbus_fact = to_dbus_fact(fact)
        return_fact = from_dbus_fact(dbus_fact)
        self.assertEqual(return_fact, fact)

        range, __ = dt.Range.parse("2020-01-19 11:00 - 2020-01-19 12:00")
        dbus_range = to_dbus_range(range)
        return_range = from_dbus_range(dbus_range)
        self.assertEqual(return_range, range)


if __name__ == '__main__':
    unittest.main()
